#include <avr/eeprom.h>
#include "global.h"
#include "align.h"
#include "nv_memory.h"

extern uint8_t DAA;

void SetDAA(int32_t nFreq)
{
	if( (nFreq >= 87500) && (nFreq <= 108000) ) {
	    DAA = eeprom_read_byte( (uint8_t*)(NVMADDR_DAA + (nFreq - 87450) / 100) );
	    if(DAA <= 127)
	        return;
	}

	Align(nFreq);
}


void WriteDAA(int32_t nFreq)
{
	if( (nFreq >= 87500) && (nFreq <= 108000) ) {
	    eeprom_write_byte( (uint8_t*)(NVMADDR_DAA + (nFreq - 87450) / 100), DAA );
	}
}


void Align(int32_t freq)
{
    // alignment of the antenna input circuit
    // these values are individual for each tuner!

    if(freq >= 181000)
        DAA = 0x00;
    else if(freq >= 108100)
        DAA = 0xB5 - (uint8_t)(freq / 1000);

    // FM data begin
    else if(freq >= 106575)
        DAA = 0x49;
    else if(freq >= 105125)
        DAA = 0x4A;
    else if(freq >= 103725)
        DAA = 0x4B;
    else if(freq >= 102275)
        DAA = 0x4C;
    else if(freq >= 100875)
        DAA = 0x4D;
    else if(freq >= 99425)
        DAA = 0x4E;
    else if(freq >= 98025)
        DAA = 0x4F;
    else if(freq >= 94525)
        DAA = 0x50;
    else if(freq >= 91025)
        DAA = 0x51;
    else if(freq >= 87525)
        DAA = 0x52;
    else if(freq >= 87500)
        DAA = 0x53;
    // FM data end

    else
        DAA = 0x68 - (uint8_t)(freq / 4000);
}
