#include "global.h"

#define NVMSIG0     'A'
#define NVMSIG1     'C'
#define NVMSIG2     0X00
#define NVMSIG3     0X00
#define NVMSIG      (NVMSIG0 + ((uint16_t)NVMSIG1 << 8) + ((uint32_t)NVMSIG2 << 16) + ((uint32_t)NVMSIG3 << 24))

// Channels in band
#define CHS_LW       10
#define CHS_MW       25
#define CHS_SW      100
#define CHS_SH       25
#define CHS_FL       25
#define CHS_FM      100
#define CHS_AL       10
#define CHS_AB       25
#define CHS_VH       25

// NV memory addr
#define NVMADDR_SIG           0
#define NVMADDR_LW            4
#define NVMADDR_MW           24
#define NVMADDR_SW           74
#define NVMADDR_SH          274
#define NVMADDR_FL          324
#define NVMADDR_FM          374
#define NVMADDR_AL          574
#define NVMADDR_AB          594
#define NVMADDR_VH          644
#define NVMADDR_BANDCH      694
#define NVMADDR_BANDFREQ    703
#define NVMADDR_DAA         721
#define NVMADDR_BAND        927
#define NVMADDR_VOL         928
#define NVMADDR_MODE        929
#define NVMADDR_RFMODE      930
#define NVMADDR_TUNE	    931
#define NVMADDR_STEPIDX     932
#define NVMADDR_AMFILTER    933
#define NVMADDR_FMFILTER    934
#define NVMADDR_MISC1	    935
#define NVMADDR_SQU1        936
#define NVMADDR_SQU2	    937
#define NVMADDR_SCSTAY	    938
#define NVMADDR_ANYHOLD	    939
#define NVMADDR_BKADJ	    940
#define NVMADDR_BKKEEP	    941


// NV memory initialization data
const uint8_t NVM_INIT[] PGM =
{
	4, NVMSIG0, NVMSIG1, NVMSIG2, NVMSIG3,

	0x80 + 20,  0xff,   // LW, ch 0-9
	0x80 + 50,  0xff,   // MW, ch 0-24
	0x80 + 100, 0xff,   // SW, ch 0-49
	0x80 + 100, 0xff,   // SW, ch 50-99
	0x80 + 50,  0xff,   // SW-H, ch 0-24
	0x80 + 50,  0xff,   // FM-L, ch 0-24
	0x80 + 100, 0xff,   // FM, ch 0-49
	0x80 + 100, 0xff,   // FM, ch 50-99
	0x80 + 20,  0xff,   // AIR-L, ch 0-9
	0x80 + 50,  0xff,   // AIR, ch 0-24
	0x80 + 50,  0xff,   // VHF, ch 0-24
	0x80 + 9,      0,   // Band ch No.
	0x80 + 18,  0xff,   // Band frequency
	0x80 + 103, 0xff,   // DAA for band FM(87.5-108M)
	0x80 + 103, 0xff,   // DAA for band FM(87.5-108M)

	8, 5, 30, 1, 1, 0, 0, 12, 29,  // Band, Vol, Mode, RFMode, TuneType, StepIdx, AMFilter, FMFilter
	7, B01110010, -30, 30, 5, 3, 255, 0,  // Misc1, Squ1, Squ2, TScan, TAny, BkAdj, BkKeep

	0                   // End
};


void NVMInit(void);
void NVMGetArgs(void);
void NVMUnpkWrData(const uint8_t* pPGMAddr, uint8_t *pNVMAddr);

#if DEBUG
uint8_t NV_read_byte(uint8_t *p);
uint16_t NV_read_word(uint16_t *p);
void NV_write_byte(uint8_t *p, uint8_t val);
void NV_write_word(uint16_t *p, uint16_t val);
#else
#define NV_read_byte(p)         eeprom_read_byte((p))
#define NV_read_word(p)         eeprom_read_word((p))
#define NV_write_byte(p, val)   eeprom_write_byte((p), (val))
#define  NV_write_word(p, val)  eeprom_write_word((p), (val))
#endif
