#include <avr/eeprom.h>
#include "global.h"
#include "ui.h"
#include "yacrc3167_ar.h"
#include "nv_memory.h"
#include "filters.h"
#include "align.h"


extern uint8_t DAA;
extern uint8_t BAND;
extern uint8_t AGC;
extern uint8_t nAGC;


extern uint8_t nMode;    // FM/AM/AUX
extern uint8_t nRFMode;  // FM/AM

extern uint8_t nVolume;    // Audio volume control, 0-63

extern uint8_t nTuneType;  // Freq_Step/Seek/Ch/Scan/Any/Scan_Save

extern uint8_t nBand;  // Band: LW/MW/SW/SH/FL/FM/AL/AB/VH
extern const int32_t nBandFMin[9];
extern const int32_t nBandFMax[9];
extern int32_t nBandFreq[9];  // Band default frequency
extern int8_t nBandCh[9];
extern const int8_t nBandChs[9];  // Band total channels
extern const int16_t nBandStep[9][3];

extern uint8_t nStepIdx;  // Step index for current band

extern int8_t nAMFilter;  // Current FIR filter index for AM
extern int8_t nFMFilter;  // Current FIR filter index for FM(ADP_FM_FILTER is adaptive)
extern uint8_t nDeemphasis;
extern int8_t nSquelch[];  // Signal squelch value in dBuv, -99~99
                           // 1st element for auto mute, 2nd element for seek/scan/any

extern uint8_t nBacklightAdj;     // LCD backlight value, 0-255
extern uint8_t nBacklightKeep;    // LCD backlight auto keep seconds, 0-255, 0 for always on
extern bool bLCDOff;              // true for LCD is off
extern uint32_t nBacklightTimer;  // LCD backlight auto keep timer in ms
extern int32_t nSecondsOffset;    // Seconds of real time offset, preset to 15 days

extern uint8_t nScanStayTime;  // Seconds to stay at current frequency
extern uint8_t nAnyHoldTime;   // Seconds to hold current frequency when lost signal


volatile int8_t nLRot;
volatile int8_t nRRot;

const char *pszTuneTypes[] = {"FS", "SK", "CH", "SC", "AN", "SS"};
const char *pszBands[] = {"LW", "MW", "SW", "SH", "FL",
                          "FM", "AL", "AB", "VH"};  // Band name to display

bool bExitMenu;
bool bInAlig = false;  // In DAA alignment routine, reserve 2nd row of LCD




// Special menu strings
const char MT_RET[] PGM = "RET";
const char MT_EXIT[] PGM = "EXIT";

// Menu Option
const char MT_SQU1[] PGM = "SQU1";
const char MT_SQU2[] PGM = "SQU2";
const char MT_GAIN[] PGM = "GAIN";
const char MT_AGC[]  PGM = "AGC";
const char MT_DEEM[] PGM = "DEEM";
const char MT_BKLT[] PGM = "BKLT";
const char MT_TSCN[] PGM = "TSCN";
const char MT_TANY[] PGM = "TANY";
const char MT_TIME[] PGM = "TIME";

M_ITEM M_Option[] = {
    {MID_SQUELCH1, MT_SQU1},
    {MID_SQUELCH2, MT_SQU2},
    {MID_GAIN, MT_GAIN},
    {MID_AGC, MT_AGC,},
    {MID_DEEM, MT_DEEM},
    {MID_BKLT, MT_BKLT},
    {MID_TSCN, MT_TSCN},
    {MID_TANY, MT_TANY},
    {MID_TIME, MT_TIME},
    {MID_EXIT, MT_EXIT}
};

// Menu Option->GAIN
const char MT_RFP0[] PGM = "NORM";
const char MT_RFP6[] PGM = "RF+6";

M_ITEM M_Gain[] = {
    {MID_RFP0, MT_RFP0},
    {MID_RFP6, MT_RFP6}
};

// Menu Option->DEEM
const char MT_DEEM0[] PGM = "OFF";
const char MT_DEEM50[] PGM = "50US";
const char MT_DEEM75[] PGM = "75US";

M_ITEM M_Deem[] = {
    {MID_DEEM0, MT_DEEM0},
    {MID_DEEM50, MT_DEEM50},
    {MID_DEEM75, MT_DEEM75}
};

// Menu Option->BKLT
const char MT_BKKEEP[] PGM = "KEEP";
const char MT_BKADJ[] PGM = "ADJ";

M_ITEM M_BkLt[] = {
    {MID_BKKEEP, MT_BKKEEP},
    {MID_BKADJ, MT_BKADJ},
    {MID_RET, MT_RET}
};



// Menu Frequency
const char MT_MODE[] PGM = "MODE";
const char MT_TUNE[] PGM = "TUNE";
const char MT_BAND[] PGM = "BAND";
const char MT_FILT[] PGM = "FILT";
const char MT_ALIGNMENT[] PGM = "ALIG";

M_ITEM M_Frequency[] = {
    {MID_MODE, MT_MODE},
    {MID_TUNE, MT_TUNE},
    {MID_BAND, MT_BAND},
    {MID_FILT, MT_FILT},
    {MID_ALIGNMENT, MT_ALIGNMENT},
    {MID_EXIT, MT_EXIT}
};


// Menu Frequency->MODE
const char MT_MODEAM[] PGM = "AM";
const char MT_MODEFM[] PGM = "FM";
const char MT_MODEAUX[] PGM = "AUX";

M_ITEM M_Mode[] = {
    {MID_MODEAM, MT_MODEAM},
    {MID_MODEFM, MT_MODEFM},
    {MID_MODEAUX, MT_MODEAUX}
};


// Menu Frequency->TUNE
const char MT_FREQ[] PGM = "FREQ";
const char MT_SEEK[] PGM = "SEEK";
const char MT_CH[] PGM = "CH";
const char MT_SCAN[] PGM = "SCAN";
const char MT_ANY[] PGM = "ANY";
const char MT_SCSV[] PGM = "S&S";

M_ITEM M_Tune[] = {
    {MID_FREQ, MT_FREQ},
    {MID_SEEK, MT_SEEK},
    {MID_CH, MT_CH},
    {MID_SCAN, MT_SCAN},
    {MID_ANY, MT_ANY},
    {MID_SCSV, MT_SCSV}
};


// Menu Frequency->BAND
const char MT_LW[] PGM = "LW";
const char MT_MW[] PGM = "MW";
const char MT_SW[] PGM = "SW";
const char MT_SH[] PGM = "SW-H";
const char MT_FL[] PGM = "FM-L";
const char MT_FM[] PGM = "FM";
const char MT_AL[] PGM = "AB-L";
const char MT_AB[] PGM = "AIR";
const char MT_VH[] PGM = "VHF";

M_ITEM M_Band[] = {
    {MID_LW, MT_LW},
    {MID_MW, MT_MW},
    {MID_SW, MT_SW},
    {MID_SH, MT_SH},
    {MID_FL, MT_FL},
    {MID_FM, MT_FM},
    {MID_AL, MT_AL},
    {MID_AB, MT_AB},
    {MID_VH, MT_VH}
};


// Menu Frequency->FILT
const char MT_FTAM00[] PGM = "1K1";
const char MT_FTAM01[] PGM = "1K9";
const char MT_FTAM02[] PGM = "2K2";
const char MT_FTAM03[] PGM = "2K5";
const char MT_FTAM04[] PGM = "2K9";
const char MT_FTAM05[] PGM = "3K4";
const char MT_FTAM06[] PGM = "3K9";
const char MT_FTAM07[] PGM = "4K6";
const char MT_FTAM08[] PGM = "5K2";
const char MT_FTAM09[] PGM = "6K0";
const char MT_FTAM10[] PGM = "6K9";
const char MT_FTAM11[] PGM = "7K9";
const char MT_FTAM12[] PGM = "9K1";
const char MT_FTAM13[] PGM = "10K4";
const char MT_FTAM14[] PGM = "11K3";
const char MT_FTAM15[] PGM = "11K8";
const char MT_FTAM16[] PGM = "13K5";
const char MT_FTAM17[] PGM = "15K6";
const char MT_FTAM18[] PGM = "17K8";
const char MT_FTAM19[] PGM = "19K9";
const char MT_FTAM20[] PGM = "22K1";
const char MT_FTAM21[] PGM = "24K3";
const char MT_FTAM22[] PGM = "26K4";
const char MT_FTAM23[] PGM = "28K6";
const char MT_FTAM24[] PGM = "30K8";
const char MT_FTAM25[] PGM = "32K9";
const char MT_FTAM26[] PGM = "35K1";
const char MT_FTAM27[] PGM = "37K3";
const char MT_FTAM28[] PGM = "38K6";

const char MT_FTFM00[] PGM = "9K";
const char MT_FTFM01[] PGM = "15K";
const char MT_FTFM02[] PGM = "17K";
const char MT_FTFM03[] PGM = "20K";
const char MT_FTFM04[] PGM = "24K";
const char MT_FTFM05[] PGM = "27K";
const char MT_FTFM06[] PGM = "32K";
const char MT_FTFM07[] PGM = "36K";
const char MT_FTFM08[] PGM = "42K";
const char MT_FTFM09[] PGM = "48K";
const char MT_FTFM10[] PGM = "55K";
const char MT_FTFM11[] PGM = "63K";
const char MT_FTFM12[] PGM = "73K";
const char MT_FTFM13[] PGM = "83K";
const char MT_FTFM14[] PGM = "90K";
const char MT_FTFM15[] PGM = "95K";
const char MT_FTFM16[] PGM = "108K";
const char MT_FTFM17[] PGM = "125K";
const char MT_FTFM18[] PGM = "142K";
const char MT_FTFM19[] PGM = "159K";
const char MT_FTFM20[] PGM = "177K";
const char MT_FTFM21[] PGM = "194K";
const char MT_FTFM22[] PGM = "211";
const char MT_FTFM23[] PGM = "229K";
const char MT_FTFM24[] PGM = "246K";
const char MT_FTFM25[] PGM = "263K";
const char MT_FTFM26[] PGM = "281K";
const char MT_FTFM27[] PGM = "298K";
const char MT_FTFM28[] PGM = "309K";
const char MT_FTFM29[] PGM = "AUTO";

M_ITEM M_AMFilter[] = {
    {MID_FTAM00, MT_FTAM00},
    {MID_FTAM01, MT_FTAM01},
    {MID_FTAM02, MT_FTAM02},
    {MID_FTAM03, MT_FTAM03},
    {MID_FTAM04, MT_FTAM04},
    {MID_FTAM05, MT_FTAM05},
    {MID_FTAM06, MT_FTAM06},
    {MID_FTAM07, MT_FTAM07},
    {MID_FTAM08, MT_FTAM08},
    {MID_FTAM09, MT_FTAM09},
    {MID_FTAM10, MT_FTAM10},
    {MID_FTAM11, MT_FTAM11},
    {MID_FTAM12, MT_FTAM12},
    {MID_FTAM13, MT_FTAM13},
    {MID_FTAM14, MT_FTAM14},
    {MID_FTAM15, MT_FTAM15},
    {MID_FTAM16, MT_FTAM16},
    {MID_FTAM17, MT_FTAM17},
    {MID_FTAM18, MT_FTAM18},
    {MID_FTAM19, MT_FTAM19},
    {MID_FTAM20, MT_FTAM20},
    {MID_FTAM21, MT_FTAM21},
    {MID_FTAM22, MT_FTAM22},
    {MID_FTAM23, MT_FTAM23},
    {MID_FTAM24, MT_FTAM24},
    {MID_FTAM25, MT_FTAM25},
    {MID_FTAM26, MT_FTAM26},
    {MID_FTAM27, MT_FTAM27},
    {MID_FTAM28, MT_FTAM28}
};

M_ITEM M_FMFilter[] = {
    {MID_FTFM00, MT_FTFM00},
    {MID_FTFM01, MT_FTFM01},
    {MID_FTFM02, MT_FTFM02},
    {MID_FTFM03, MT_FTFM03},
    {MID_FTFM04, MT_FTFM04},
    {MID_FTFM05, MT_FTFM05},
    {MID_FTFM06, MT_FTFM06},
    {MID_FTFM07, MT_FTFM07},
    {MID_FTFM08, MT_FTFM08},
    {MID_FTFM09, MT_FTFM09},
    {MID_FTFM10, MT_FTFM10},
    {MID_FTFM11, MT_FTFM11},
    {MID_FTFM12, MT_FTFM12},
    {MID_FTFM13, MT_FTFM13},
    {MID_FTFM14, MT_FTFM14},
    {MID_FTFM15, MT_FTFM15},
    {MID_FTFM16, MT_FTFM16},
    {MID_FTFM17, MT_FTFM17},
    {MID_FTFM18, MT_FTFM18},
    {MID_FTFM19, MT_FTFM19},
    {MID_FTFM20, MT_FTFM20},
    {MID_FTFM21, MT_FTFM21},
    {MID_FTFM22, MT_FTFM22},
    {MID_FTFM23, MT_FTFM23},
    {MID_FTFM24, MT_FTFM24},
    {MID_FTFM25, MT_FTFM25},
    {MID_FTFM26, MT_FTFM26},
    {MID_FTFM27, MT_FTFM27},
    {MID_FTFM28, MT_FTFM28},
    {MID_FTFM29, MT_FTFM29}
};


M_SUBMENU SM_List[] = {
    {MID_OPTION, M_Option, sizeof(M_Option) / sizeof(M_ITEM)},  // Menu Option
    {MID_FREQUENCY, M_Frequency, sizeof(M_Frequency) / sizeof(M_ITEM)},  //  Menu Frequency
    {MID_BKLT, M_BkLt, sizeof(M_BkLt) / sizeof(M_ITEM)},  // Menu Option->BKLT
    {MID_GAIN, M_Gain, sizeof(M_Gain) / sizeof(M_ITEM)},  // Menu Option->GAIN
    {MID_DEEM, M_Deem, sizeof(M_Deem) / sizeof(M_ITEM)},  // Menu Option->DEEM
    {MID_MODE, M_Mode, sizeof(M_Mode) / sizeof(M_ITEM)},  // Menu Frequency->MODE
    {MID_TUNE, M_Tune, sizeof(M_Tune) / sizeof(M_ITEM)},  // Menu Frequency->TUNE
    {MID_BAND, M_Band, sizeof(M_Band) / sizeof(M_ITEM)},  // Menu Frequency->BAND
    {MID_FILT, M_FMFilter, sizeof(M_FMFilter) / sizeof(M_ITEM)}  // Menu Frequency->FILT, toggle AM/FM by nRFMode
};




////////////////////////////////////////////////////////////
// LCD utility
////////////////////////////////////////////////////////////
void LCDEN(void)  // Generate LCD EN pulse
{
  digitalWrite(LCD_E_PIN, HIGH);
  delayMicroseconds(1);
  digitalWrite(LCD_E_PIN, LOW);
  //delayMicroseconds(1);
}


void LCDCmd(uint8_t Cmd)  // Write LCD command
{
  digitalWrite(LCD_RS_PIN, LOW);
  PORTB = Cmd >> 4;  // High 4 bits, with LCD E pin low
  LCDEN();
  PORTB = Cmd & 0x0f;  // Low 4 bits, with LCD E pin low
  LCDEN();
  delayMicroseconds(50);
}


void LCDData(uint8_t Data)  // Write LCD data
{
  digitalWrite(LCD_RS_PIN, HIGH);
  PORTB = (Data >> 4);  // High 4 bits, with LCD E pin low
  LCDEN();
  PORTB = (Data & 0x0f);  // Low 4 bits, with LCD E pinlow
  LCDEN();
  delayMicroseconds(50);
}


void LCDClr(void)
{
    LCDCmd(0x01);  // Clear LCD
    delay(5);
}


void LCDClr1(void)
{
    LCDXYStr_P(0, 1, (PGM_P)F("                "));  // Clear LCD's 2nd row
}


void LCDOn(void)
{
    if(nBacklightKeep && bLCDOff) {
        analogWrite(LCD_BKLT_PIN, nBacklightAdj);  // Set LCD backlight PWM
        bLCDOff = false;
    }
    nBacklightTimer = millis();
}


void LCDInit(void)
{
    DDRB = B111111;  // D8-D13 output
    pinMode(LCD_RS_PIN, OUTPUT);
    analogWrite(LCD_BKLT_PIN, 255);  // Set LCD backlight PWM always on for booting

    LCDCmd(0x28);  // 4 bits interface
    delay(5);
    LCDCmd(0x28);
    delay(5);
    LCDCmd(0x28);
    delay(5);

    LCDCmd(0x08);  // LCD off
    delayMicroseconds(50);
    LCDClr();      // Clear LCD
    LCDCmd(0x06);  // Auto increment
    delayMicroseconds(50);
    LCDCmd(0x0c);  // LCD on
    delayMicroseconds(50);
}


void LCDXY(uint8_t x, uint8_t y)  // x:0-15, y:0-1
{
    LCDCmd(0x80 + y * 0x40 + x);
}


void LCDXYChar(uint8_t x, uint8_t y, char c)  // Display char at x:0-15, y:0-1
{
    LCDXY(x, y);
    LCDData(c);
}


void LCDXYStr(uint8_t x, uint8_t y, const char *str)  // Display string at x:0-15, y:0-1
{
    const char *p;
    char c;

    LCDXY(x, y);
    p = str;
    while( (c = *p++) )
        LCDData(c);
}


void LCDXYStr_P(uint8_t x, uint8_t y, PGM_P str)  // Display PROGMEM string at x:0-15, y:0-1
{
    const char *p;
    char c;

    LCDXY(x, y);
    p = str;
    while( (c = pgm_read_byte_near(p++)) )
        LCDData(c);
}


// Display string at x:0-15, y:0-1
void LCDXYStrLen(uint8_t x, uint8_t y, char *str, uint8_t nLen, bool bLeftAlign)
{
    char *p, c;
    uint8_t i, n;

    LCDXY(x, y);
    p = str;
    n = nLen - strlen(str);
    if(!bLeftAlign) {
        for(i = 0; i < n; i++)
            LCDData(' ');
    }

    while( (c = *p++) )
        LCDData(c);

    if(bLeftAlign) {
        for(i = 0; i < n; i++)
            LCDData(' ');
    }
}


// Display PROGMEM string at x:0-15, y:0-1
void LCDXYStrLen_P(uint8_t x, uint8_t y, PGM_P str, uint8_t nLen, bool bLeftAlign)
{
    char c;
    const char *p;
    uint8_t i, n;

    LCDXY(x, y);
    p = str;
    n = nLen - strlen_P(str);
    if(!bLeftAlign) {
        for(i = 0; i < n; i++)
            LCDData(' ');
    }

    while( (c = pgm_read_byte_near(p++)) )
        LCDData(c);

    if(bLeftAlign) {
        for(i = 0; i < n; i++)
            LCDData(' ');
    }
}


// Display signed integer at x:0-15, y:0-1. Righ aligned, spaces filled at left. Too long integers are not allowed.
void LCDXYIntLen(uint8_t x, uint8_t y, int32_t n, uint8_t nLen)
{
    int8_t i;
    int32_t r;

    if(n >= 0)
        r = n;
    else
        r = -n;

    i = nLen - 1;
    do {
        LCDXYChar(x + i--, y, '0' + (r % 10));
        r = r / 10;
    } while(r);

    if(n < 0)
        LCDXYChar(x + i--, y, '-');

    while(i >= 0)
        LCDXYChar(x + i--, y, ' ');
}


// Display unsigned integer at x:0-15, y:0-1. Righ aligned, '0' filled at left. Too long unsigned integers are not allowed.
void LCDXYUIntLenZP(uint8_t x, uint8_t y, uint32_t n, uint8_t nLen)
{
    int8_t i;
    uint32_t r;

    r = n;
    i = nLen - 1;
    do {
        LCDXYChar(x + i--, y, '0' + (r % 10));
        r = r / 10;
    } while(r);

    while(i >= 0)
        LCDXYChar(x + i--, y, '0');
}


////////////////////////////////////////////////////////////
// Rotary encoder & key utility
////////////////////////////////////////////////////////////
void intLRotEnc(void)
{
    if( digitalRead(SH1A_PIN) == digitalRead(SH1B_PIN) )
        ++nLRot;
    else
        --nLRot;
}


int8_t GetLRot(void)
{
    int8_t n;

    if(nLRot) {
        noInterrupts();
        n = nLRot;
        nLRot = 0;
        interrupts();
        LCDOn();
        return n;
    }
    return 0;
}


void intRRotEnc(void)
{
    if( digitalRead(SH2A_PIN) == digitalRead(SH2B_PIN) )
        ++nRRot;
    else
        --nRRot;
}


int8_t GetRRot(void)
{
    int8_t n;

    if(nRRot) {
        noInterrupts();
        n = nRRot;
        nRRot = 0;
        interrupts();
        LCDOn();
        return n;
    }
    return 0;
}


uint8_t PeekKey(void)
{
    int8_t i;
    uint16_t nADC;
    uint8_t nKey0 = 0xff, nKey;

    for(i = 0; i < 20;) {
        nADC = analogRead(0);
        if( nADC < (ADC_LROT + ADC_RROT) / 2 )
            nKey = KEY_LROT;
        else if( nADC < (ADC_RROT + ADC_TUNE) / 2 )
            nKey = KEY_RROT;
        else if( nADC < (ADC_TUNE + ADC_STEP) / 2 )
            nKey = KEY_TUNE;
        else if(nADC < (ADC_STEP + ADC_BAND) / 2 )
            nKey = KEY_STEP;
        else if( nADC < (ADC_BAND + ADC_FILTER) / 2 )
            nKey = KEY_BAND;
        else if( nADC < (ADC_FILTER + ADC_NONE) / 2 )
            nKey = KEY_FILTER;
        else
            nKey = 0;

        if(nKey == nKey0)
          ++i;
        else {
            i = 0;
            nKey0 = nKey;
        }
    }
    return nKey;
}


uint8_t GetKey(void)
{
    uint8_t nKey0, nKey;
    uint32_t timer1;  // Long press timer
    static uint8_t nKeyWaitUp = 0;
    static uint32_t tKeyWaitUp;

    if( nKeyWaitUp && (millis() - tKeyWaitUp) < TIMER_LAST_LP )
        while( ( nKey0 = PeekKey() ) == nKeyWaitUp )
            ;
    else
        nKey0 = PeekKey();

    nKeyWaitUp = 0;
    if( !nKey0 )
        return 0;

    timer1 = millis();
    while( (millis() - timer1) < TIMER_LONGPRESS ) {
        if( !( nKey = PeekKey() ) ) {
            LCDOn();
#if DEBUG
            Serial.print(nKey0, HEX);
            Serial.println(F(" Key up"));
#endif
            return nKey0;  // Key up
        }

        if( nKey != nKey0 ) {  // Another key pressed
            nKey0 = nKey;
            timer1 = millis();  // Reset long press timer for new key
            continue;
        }
    }

    tKeyWaitUp = millis();
    nKeyWaitUp = nKey0;
    LCDOn();
#if DEBUG
    Serial.print(nKey0, HEX);
    Serial.println(F(" Key LP"));
#endif
    return nKey0 | KEY_LONGPRESS;
}


void TestRotKey(void)  // Show Rotary encoder and key data on LCD
{
    uint8_t uLRot = 0;
    int8_t iRRot = 0, i, iTmp;

    LCDClr();
    LCDXYStr_P(6, 0, (PGM_P)F("TEST"));
    nLRot = 0;
    nRRot = 0;

    for(i = 0; ; i++) {
        iTmp = GetLRot();
        if(iTmp && PeekKey() == KEY_RROT) {  // Right rotary encoder pressed while rotate left rotary encoder
            if( (iTmp < 0) && (uLRot == 67) && (iRRot == 77) )  // Rotate CCW
                NVMInit();  // Reinitialize NV memory
            return;
        }
        uLRot += iTmp;
        LCDXYIntLen(0, 0, uLRot, 4);

        iRRot += GetRRot();
        LCDXYIntLen(12, 0, iRRot, 4);

        if( !(i % 25) )
            LCDXYIntLen(6, 1, analogRead(0), 4);
    }
}


////////////////////////////////////////////////////////////
// User interface
////////////////////////////////////////////////////////////
void LCDUpdate(void)
{
    int8_t i8;
    PGM_P p;
    const char cmode[] = "AFX";

    // Update mode: A/F/X, A=AM, F=FM, X=AUX
    LCDXYChar(MODE_X, MODE_Y, cmode[nMode]);

    // Update band: LW/MW/SW/SH/FL/FM/AL/AB/VH
    // SH=SW high, FL=FM Low, AL=Air band Low, AB=Air Band, VH=VHF
    LCDXYStr(BAND_X, BAND_Y, (char *)pszBands[nBand]);

    if(!bInAlig) {
        // Update tune type: FS/SK/CH/SC/AN/SS
            LCDXYStr(TYPE_X, TYPE_Y, (char *)pszTuneTypes[nTuneType]);

        if( (nTuneType == TYPE_CH) || (nTuneType == TYPE_SCSV) )  // CH/SS type
            LCDXYUIntLenZP(STEP_X, STEP_Y, nBandCh[nBand], 3);
        else {  // None ch/ss type
            // Update step: 1K/5K/9K/10K/25K/45K/50K/90K/100/500
            if(nBandStep[nBand][nStepIdx] < 100)
                i8 = 1;
            else
                i8 = 0;
            LCDXYIntLen(STEP_X, STEP_Y, nBandStep[nBand][nStepIdx], 3 - i8);
            if(i8)
                LCDXYChar(STEP_X + 2, STEP_Y, 'K');
        }

        // Update FIR filter bandwidth
        if(nRFMode == MODE_FM) {
            p = M_FMFilter[nFMFilter].pszMTxt;
        }
        else
            p = M_AMFilter[nAMFilter].pszMTxt;
        LCDXYStrLen_P(FILTER_X, FILTER_Y, p, 4, false);
    }
}


int8_t CheckUpdateSig(void)
{
    int8_t i8;

    if(nRFMode == MODE_FM && dsp_query1(DSP_ST_19kHz))
        i8 = 'S';  // Stereo
    else
        i8 = ' ';  // Mono

    // 19kHz stereo subcarrier indicator, 'S' for stereo, ' ' for mono
    LCDXYChar(STEREO_X, STEREO_Y, i8);  // Update FM stereo indicator

    i8 = GetdBuv();
    LCDXYIntLen(RSSI_X, RSSI_Y, constrain(i8, -99, 99), 3);  // Update RF signal level in dBuv
    return i8;
}


void ShowMisc(void)
{
    char s[] = "3Ri- ";
    const char ds[] = "-57";

    if(nRFMode == MODE_FM)
        s[3] = ds[nDeemphasis];  // Deemphasis, '-' for none, '5' for 50us, '7' for 75us

    s[0] = '0' + nAGC;

    if( !(AGC & 0x1) || (BAND & 0x01) )
        s[1] = 'r';  // FM mixer normal gain, or AM mode

    LCDXYStr(ALT_X, ALT_Y, s);
}


void ShowTime(void)
{
    uint32_t nMinutes;

    nMinutes = (millis() / 1000 + nSecondsOffset) / 60;
    LCDXYUIntLenZP(ALT_X, ALT_Y, (nMinutes / 60) % 24, 2);  // Hour
    LCDXYChar(ALT_X + 2, ALT_Y, ':');
    LCDXYUIntLenZP(ALT_X + 3, ALT_Y, nMinutes % 60, 2);  // Minute
}


void ShowVol(void)
{
    LCDXYStr_P(ALT_X, ALT_Y, (PGM_P)F("VOL"));
    LCDXYIntLen(ALT_X + 3, ALT_Y, nVolume, 2);  // Volume
}


void CheckUpdateAlt(int8_t nShow)  // Check and update ALT area
{
    static int8_t nShowLast = ALT_MISC;
    static uint32_t nTimerAlt = 0;

    if(nShow != ALT_AUTO) {
        nTimerAlt = millis();
        nShowLast = nShow;
    }

    if( (millis() - nTimerAlt) < TIMER_ALTDISP ) {  // Not timeout
        switch(nShowLast) {
        case ALT_MISC:
            ShowMisc();
            break;

        case ALT_TIME:
            ShowTime();
            break;

        case ALT_VOL:
            ShowVol();
            break;
        }
    }
    else {  // Timeout, swap beteen ALT_MISC/ALT_TIME
        if(nShowLast != ALT_MISC) {
            nShowLast = ALT_MISC;
            ShowMisc();
        }
        else {
            nShowLast = ALT_TIME;
            ShowTime();
        }
        nTimerAlt = millis();
    }
}


void Menu_Squelch(uint8_t nIdx)
{
    int16_t i16 = nSquelch[nIdx];
    uint8_t nKey;

                            // 0123456789012345
    LCDXYStr_P(0, 1, (PGM_P)F("SQUELCH1:       "));
    if(nIdx)
        LCDXYChar(7, 1, '2');

    for(;;) {
        i16 += GetLRot() + GetRRot();
        i16 = constrain(i16, -99, 99);
        LCDXYIntLen(10, 1, i16, 3);

        if( (nKey = GetKey()) ) {
            nSquelch[nIdx] = (int8_t)i16;
            LCDClr1();
            if( !(nKey & (KEY_LROT | KEY_RROT)) )
                bExitMenu = true;
            NV_write_byte((uint8_t *)NVMADDR_SQU1 + nIdx, nSquelch[nIdx]);
            return;
        }

        CheckUpdateSig();
        delay(100);
    }

}


void Menu_AGC(void)
{
    int16_t i16;
    uint8_t nKey;

                            // 0123456789012345
    LCDXYStr_P(0, 1, (PGM_P)F("AGC THRESHOLD:  "));

    i16 = nAGC;

    for(;;) {
        i16 += GetLRot() + GetRRot();
        i16 = constrain(i16, 0, 3);
        LCDXYIntLen(15, 1, i16, 1);

        if( (nKey = GetKey()) ) {
            nAGC = (uint8_t)i16;
            LCDClr1();
            if( !(nKey & (KEY_LROT | KEY_RROT)) )
                bExitMenu = true;
            TuneReg();
            AddSyncBits(NEEDSYNC_MISC1);
            return;
        }

        CheckUpdateSig();
        delay(100);
    }

}


void Menu_BacklightAdj(void)
{
    int16_t i16 = nBacklightAdj;
    uint8_t nKey;

                            // 0123456789012345
    LCDXYStr_P(0, 1, (PGM_P)F("BACKLT ADJ:     "));

    for(;;) {
        i16 =  (i16 + GetLRot() + GetRRot() + 256) % 256;
        LCDXYIntLen(12, 1, i16, 3);
        analogWrite(LCD_BKLT_PIN, i16);

        if( (nKey = GetKey()) ) {
            nBacklightAdj = (uint8_t)i16;
            LCDClr1();
            if( !(nKey & (KEY_LROT | KEY_RROT)) )
                bExitMenu = true;
            NV_write_byte((uint8_t *)NVMADDR_BKADJ, nBacklightAdj);
            return;
        }

        CheckUpdateSig();
        delay(100);
    }

}


void Menu_BacklightKeep(void)
{
    int16_t i16 = nBacklightKeep;
    uint8_t nKey;

                            // 0123456789012345
    LCDXYStr_P(0, 1, (PGM_P)F("BACKLT KEEP:    "));

    for(;;) {
        i16 =  (i16 + GetLRot() + GetRRot() + 256) % 256;
        LCDXYIntLen(13, 1, i16, 3);

        if( (nKey = GetKey()) ) {
            nBacklightKeep = (uint8_t)i16;
            LCDClr1();
            if( !(nKey & (KEY_LROT | KEY_RROT)) )
                bExitMenu = true;
            NV_write_byte((uint8_t *)NVMADDR_BKKEEP, nBacklightKeep);
            return;
        }

        CheckUpdateSig();
        delay(100);
    }

}


void Menu_ScanStayTime(void)
{
    int16_t i16 = nScanStayTime;
    uint8_t nKey;

                            // 0123456789012345
    LCDXYStr_P(0, 1, (PGM_P)F("TIME SCAN:     S"));

    for(;;) {
        i16 += GetLRot() + GetRRot();
        i16 = constrain(i16, 0, 255);
        LCDXYIntLen(12, 1, i16, 3);

        if( (nKey = GetKey()) ) {
            nScanStayTime = (uint8_t)i16;
            LCDClr1();
            if( !(nKey & (KEY_LROT | KEY_RROT)) )
                bExitMenu = true;
            NV_write_byte((uint8_t *)NVMADDR_SCSTAY, nScanStayTime);
            return;
        }

        CheckUpdateSig();
        delay(100);
    }

}


void Menu_AnyHoldTime(void)
{
    int16_t i16 = nAnyHoldTime;
    uint8_t nKey;

                            // 0123456789012345
    LCDXYStr_P(0, 1, (PGM_P)F("TIME ANY:     S"));

    for(;;) {
        i16 += GetLRot() + GetRRot();
        i16 = constrain(i16, 0, 255);
        LCDXYIntLen(11, 1, i16, 3);

        if( (nKey = GetKey()) ) {
            nAnyHoldTime = (uint8_t)i16;
            LCDClr1();
            if( !(nKey & (KEY_LROT | KEY_RROT)) )
                bExitMenu = true;
            NV_write_byte((uint8_t *)NVMADDR_ANYHOLD, nAnyHoldTime);
            return;
        }

        CheckUpdateSig();
        delay(100);
    }

}


void Menu_Time(void)
{
    int32_t nSeconds;
    uint8_t u8;

                         // 0123456789012345
    LCDXYStr_P(0, 1, (PGM_P)F("TIME    :  :    "));

    for(;;) {
        nSecondsOffset += (int16_t)GetLRot() * 3600 + (int16_t)GetRRot() * 60;
        nSeconds = millis() / 1000 + nSecondsOffset;
        if(nSeconds < 0) {
            nSeconds = 0;
            nSecondsOffset = 3600L * 24 * 15 - millis() / 1000;
        }

        LCDXYUIntLenZP(6, 1, (nSeconds / 3600) % 24, 2);  // Hours
        LCDXYUIntLenZP(9, 1, (nSeconds / 60) % 60, 2);  // Minutes
        LCDXYUIntLenZP(12, 1, nSeconds % 60, 2);  // Seconds

        if( (u8 = GetKey()) ) {
            nSeconds = millis() / 1000 + nSecondsOffset;
            if(u8 == (KEY_FILTER | KEY_LONGPRESS)) {  // Adjust to on the minute, i.e. hh:mm:00
                u8 = nSeconds % 60;
                if(u8 < 30)
                    nSeconds -= u8;
                else
                    nSeconds += 60 - u8;
            }
            nSecondsOffset = nSeconds - millis() / 1000;
            LCDClr1();
            if( !(u8 & (KEY_LROT | KEY_RROT)) )
                bExitMenu = true;
            return;
        }

        CheckUpdateSig();
        delay(100);
    }

}


void Menu_Alignment(void)
{
    int16_t i16;
    int8_t i8;
    uint8_t nKey;
    bool bReDisp = true;


    bInAlig = true;
    for(;;) {
        if(bReDisp) {
                                    // 0123456789012345
            LCDXYStr_P(0, 1, (PGM_P)F("ALIGN:       S/R"));
            LCDXYIntLen(7, 1, DAA, 3);
            bReDisp = false;
        }

        if( (i8 = GetRRot()) ) {
            switch(nTuneType) {
            case TYPE_FREQ:
                nBandFreq[nBand] += i8 * nBandStep[nBand][nStepIdx];
                AdjFreq();
                TuneFreqDisp();
                break;

            case TYPE_SEEK:
                Seek(constrain(i8, -1, 1));
                break;

            case TYPE_CH:
                if(i8 < 0)
                    SeekCh(-1);
                else
                    SeekCh(1);
                break;
            }

            bReDisp = true;
        }

        if( (i8 = GetLRot()) ) {
            i16 = (int16_t)DAA + i8;
            DAA = (uint8_t)constrain(i16, 0, 127);
            TuneReg();
            bReDisp = true;
        }

        if( (nKey = GetKey()) ) {
            switch(nKey) {
            case KEY_LROT:  // Store DAA
                WriteDAA(nBandFreq[nBand]);
                break;

            case KEY_RROT:  // Resume default DAA
                DAA = 0xff;
                WriteDAA(nBandFreq[nBand]);
                SetDAA(nBandFreq[nBand]);
                TuneReg();
                break;

            default:
                LCDClr1();
                bExitMenu = true;
                bInAlig = false;
                return;
            }

            bReDisp = true;
        }

        CheckUpdateSig();
        delay(100);
    }
}


bool YesNo(bool bChkSig)
{
    uint8_t u8;

    LCDXYStr_P(13, 1, (PGM_P)F("Y/N"));
    GetKey();  // Clear key
    for(;;) {
        if( (u8 = GetKey()) ) {
            if(u8 == KEY_LROT)
                return true;
            else if(u8 == KEY_RROT)
                return false;
        }

        if(bChkSig)
            CheckUpdateSig();
        delay(100);
    }
}


void Menu_SCSV(void)
{
    uint16_t i;
    uint8_t nKey, j;
    bool bReturn = false;

                            // 0123456789012345
    LCDXYStr_P(0, 1, (PGM_P)F("Overwrite?   "));  // Confirm overwrite currrent band ch data
    if(!YesNo(true))
        return;

    nTuneType = TYPE_SCSV;
    SetVolume(0);  // Mute
    if(nRFMode == MODE_AM) {
        if(nAMFilter > SEEK_AM_FILTER)
            dsp_set_filter(SEEK_AM_FILTER);
    }
    else {
        if(nFMFilter > SEEK_FM_FILTER)
            dsp_set_filter(SEEK_FM_FILTER);
    }
    LCDClr1();
    LCDUpdate();

    nBandCh[nBand] = 0;
    for(i = 0; i <= ((nBandFMax[nBand] - nBandFMin[nBand]) / nBandStep[nBand][0]); i++) {
        nBandFreq[nBand] = nBandFMin[nBand] + (int32_t)nBandStep[nBand][0] * i;
        TuneFreqDisp();
        delay(5);

        for(j = 0; j < 30; j++) {
            delay(1);
            if(GetdBuv() >= nSquelch[1]) {  // Find one signal
                WriteChFreq(true);
                LCDUpdate();
                if(++nBandCh[nBand] >= nBandChs[nBand])  // NV memory full for current band
                    bReturn = true;
                break;
            }
        }

        CheckUpdateSig();
        CheckUpdateAlt(ALT_AUTO);

        if(bReturn)
            break;
        if( (nKey = GetKey()) )
            break;
    }

    for(; nBandCh[nBand] < nBandChs[nBand]; ++nBandCh[nBand])
        WriteChFreq(false);  // Delete extra band chs

    SetFilter(false);
    nTuneType = TYPE_CH;
    nBandCh[nBand] = 0;
    SeekCh(0);
    AddSyncBits(NEEDSYNC_TUNE);
    SetVolume(nVolume);  // Unmute
}


void AddDelCh(void)
{
    uint8_t u8;
    int8_t i8;
    bool bReDISP = true;
    bool bReturn;
    uint32_t u32 = 0;
    char s[4] = "A/D";

                            // 0123456789012345
    LCDXYStr_P(0, 1, (PGM_P)F("CH           "));
    for(bReturn = false; !bReturn; ) {
        if( (i8 = GetLRot()) ) {
            nBandCh[nBand] = (nBandCh[nBand] + i8 + nBandChs[nBand]) % nBandChs[nBand];
            bReDISP = true;
        }

        if( (i8 = GetRRot()) ) {
            nBandFreq[nBand] += i8 * nBandStep[nBand][nStepIdx];
            AdjFreq();
            TuneFreqDisp();
        }

        if(bReDISP) {
            u32 = ReadChFreq();
            LCDXYUIntLenZP(2, 1, nBandCh[nBand], 3);
            LCDXYIntLen(6, 1, u32, 6);
            if(u32)  // Has frequency
                s[2] = 'D';  // Not empty ch, can be deleted
            else
                s[2] = 'd';  // Empty ch
            LCDXYStr(13, 1, s);
            bReDISP = false;
        }

        if( (u8 = GetKey()) ) {
            switch(u8) {
            case KEY_LROT:  // Add ch
                if(YesNo(true)) {  // Confirmed
                    WriteChFreq(true);  // Add current frequency to this ch
                    u32 = nBandFreq[nBand];
                }
                bReDISP = true;
                break;

            case KEY_RROT:  // Del ch
                if(u32) {  // Not empty ch, can be deleted
                    if(YesNo(true)) {  // Confirmed
                        WriteChFreq(false);  // Delete this ch
                        u32 = 0;
                    }
                    bReDISP = true;
                }
                break;

           case KEY_STEP:
                nStepIdx = (nStepIdx + 1) % NUM_STEPS;
               break;

            case KEY_STEP | KEY_LONGPRESS:   // Default step
                nStepIdx = 0;
                break;

            default:
                bReturn = true;
                break;
            }
        }

        CheckUpdateSig();
        delay(100);
    }

    LCDClr1();
    LCDUpdate();
    CheckUpdateAlt(ALT_AUTO);
}



////////////////////////////////////////////////////////////
// Menu
////////////////////////////////////////////////////////////


void ProcSubMenu(struct M_SUBMENU *pSubMenu)
{
    int8_t i8, nHit;
    uint8_t u8;
    int8_t nFirst = 0;

    Serial.print(pSubMenu->nMID, HEX);
    Serial.println(F("  ProcSubMenu"));

    for(;;) {

        // Show max 3 sub menu items
        LCDClr1();
        for(i8 = 0; i8 < ((pSubMenu->nItemCount < 3)? pSubMenu->nItemCount : 3); i8++)
            LCDXYStrLen_P(1 + i8 * 5, 1, (pSubMenu->pMItem + ( (nFirst + i8) % pSubMenu->nItemCount) )->pszMTxt, 4, true);

        // Display special indicator char for selected item
        switch(pSubMenu->nMID) {
        case MID_DEEM:
            nHit = nDeemphasis;
            break;

        case MID_MODE:
            nHit = nMode;
            break;

        case MID_TUNE:
            nHit = nTuneType;
            break;

        case MID_BAND:
            nHit = nBand;
            break;

        case MID_FILT:
            if(nRFMode == MODE_AM)
                nHit = nAMFilter;
            else
                nHit = nFMFilter;
            break;

        default:
            nHit = 123;  // Magic number, no selected item
            break;
        }
        if(pSubMenu->nMID == MID_GAIN) {
            if( (!nFirst && (AGC & 0x01)) || (nFirst && !(AGC & 0x01)) )
                nHit = 1;
            else
                nHit = 0;
        }
        else if(nHit != 123)
            nHit = (nHit - nFirst + pSubMenu->nItemCount) % pSubMenu->nItemCount;
        if((nHit >= 0) && (nHit <= 2))
            LCDXYChar(nHit * 5, 1, CHAR_SEL);

        for(;;) {
            // Process key
            if( (i8 = GetKey()) ) {
                if(i8 & (KEY_LROT | KEY_RROT)) {  // Item selected
                    u8 = (pSubMenu->pMItem + ((nFirst + 1) % pSubMenu->nItemCount) )->nMID;
                    if(u8 == MID_RET)
                        return;
                    ProcMenuItem(u8);
                    if((pSubMenu->nMID >= MID_MIN_AUTORET) || bExitMenu) {
                        bExitMenu = true;
                        return;  // Auto return sub menu, or none left/right key pressed in sub menu
                    }
                    else
                        break;  // Redraw sub menu items
                }
                else {
                    bExitMenu = true;  // None left/right key pressed, exit
                    return;
                }
            }

            // Process rotary encoder, adjust nFirst
            if( ( i8 = (GetLRot() + GetRRot()) ) ) {
                nFirst = (nFirst + i8 + pSubMenu->nItemCount) % pSubMenu->nItemCount;
                break;
            }

           // Update sig
            CheckUpdateSig();
            delay(100);
        }
    }
}


void ProcMenuItem(uint8_t nMenuID)
{
    Serial.print(nMenuID, HEX);
    Serial.println(F(" ProcMenuItem"));

    if(bExitMenu)
        return;

    // Sub menu items
    if(nMenuID <= MID_MAX_SUB) {
        ProcSubMenu(&SM_List[nMenuID]);
        return;
    }

    // Leaf menu items(w/o sub menu)
    if( (nMenuID >= MID_FTAM00) && (nMenuID <= MID_FTAM28) ) {
        nAMFilter = nMenuID - MID_FTAM00;
        SetFilter(true);  // Set FIR filter for AM
        return;
    }

    if( (nMenuID >= MID_FTFM00) && (nMenuID <= MID_FTFM29) ) {
        nFMFilter = nMenuID - MID_FTFM00;
        SetFilter(true);  // Set FIR filter for FM
        return;
    }

    if( (nMenuID >= MID_LW) && (nMenuID <= MID_VH) ) {
        ProcBand(BAND_LW + (nMenuID - MID_LW));  // Set band
        return;
    }

    if( (nMenuID >= MID_FREQ) && (nMenuID <= MID_ANY) ) {
        nTuneType = TYPE_FREQ + (nMenuID - MID_FREQ);  // Set tune type
        AddSyncBits(NEEDSYNC_TUNE);
        return;
    }

    if( (nMenuID >= MID_DEEM0) && (nMenuID <= MID_DEEM75) ) {
        dsp_set_deemphasis(DEEMPHOFF + (nMenuID - MID_DEEM0));  // Set deemphasis
        AddSyncBits(NEEDSYNC_MISC1);
        return;
    }


    // Misc leaf menu items(w/o sub menu)
    switch(nMenuID) {
    case MID_EXIT:
        bExitMenu = true;

    case MID_RET:
        return;

    case MID_SQUELCH1:
        Menu_Squelch(0);
        break;

    case MID_SQUELCH2:
        Menu_Squelch(1);
        break;

    case MID_RFP0:
        AGC &= 0xfe;  // FM RF normal gain
        TuneReg();
        AddSyncBits(NEEDSYNC_MISC1);
        break;

    case MID_RFP6:
        AGC |= 0x01;  // FM RF gain +6dB
        TuneReg();
        AddSyncBits(NEEDSYNC_MISC1);
        break;

    case MID_AGC:
        Menu_AGC();
        break;

    case MID_BKKEEP:
        Menu_BacklightKeep();
        break;

    case MID_BKADJ:
        Menu_BacklightAdj();
        break;

    case MID_TSCN:
        Menu_ScanStayTime();
        break;

    case MID_TANY:
        Menu_AnyHoldTime();
        break;

    case MID_TIME:
        Menu_Time();  // Adjust time
        break;

    case MID_MODEAM:
        SetModeAM();  // AM
        break;

    case MID_MODEFM:
        SetModeFM();  // FM
        break;

    case MID_MODEAUX:
        SetModeAUX();  // AUX
        break;

    case MID_ALIGNMENT:
        Menu_Alignment();
        break;

    case MID_SCSV:
        Menu_SCSV();
        break;
    }
}


void Menu(uint8_t nMenuID)
{
    bExitMenu = false;
    ProcMenuItem(nMenuID);
    LCDClr1();
    LCDUpdate();
}
