#include "global.h"


#define KEY_LROT      0x01
#define KEY_RROT      0x02
#define KEY_TUNE      0x04
#define KEY_STEP      0x08
#define KEY_BAND      0x10
#define KEY_FILTER    0x20
#define KEY_LONGPRESS 0x80

#define TIMER_LONGPRESS    1200  // Key long press time, ms
#define TIMER_LAST_LP       400  // Is last long press key? ms

// The following 7 values are individual for each device
// Calculated values
/*
#define ADC_LROT    14
#define ADC_RROT    172
#define ADC_TUNE    343
#define ADC_STEP    501
#define ADC_BAND    655
#define ADC_FILTER  812
#define ADC_NONE    1023
*/

// My device's values
#define ADC_LROT    12
#define ADC_RROT    172
#define ADC_TUNE    344
#define ADC_STEP    503
#define ADC_BAND    657
#define ADC_FILTER  811
#define ADC_NONE    996


// Char display on LCD for selected item
#define CHAR_SEL        0x7e  // Right arrow


// LCD locations
#define FREQ_X    5
#define FREQ_Y    0

#define MODE_X    0
#define MODE_Y    0

#define TYPE_X    0
#define TYPE_Y    1

#define BAND_X    2
#define BAND_Y    0

#define STEP_X    2
#define STEP_Y    1

#define FILTER_X  6
#define FILTER_Y  1

#define RSSI_X   12
#define RSSI_Y    0

#define STEREO_X 15
#define STEREO_Y  0

#define ALT_X    11
#define ALT_Y     1


// Display contents in ALT area
#define ALT_AUTO  0
#define ALT_MISC  1
#define ALT_TIME  2
#define ALT_VOL   3




// Menu IDs with sub menus, not auto return
#define MID_OPTION      0x00
#define MID_FREQUENCY   0x01
#define MID_BKLT        0x02


// Menu IDs with sub menus, auto return
#define MID_MIN_AUTORET 0x03  // Min ID number with auto return property
#define MID_GAIN        0x03
#define MID_DEEM        0x04
#define MID_MODE        0x05
#define MID_TUNE        0x06
#define MID_BAND        0x07
#define MID_FILT        0x08
#define MID_MAX_SUB     0x08  // Max ID number with sub menus

// Menu IDs of leaves
#define MID_SQUELCH1    0x20
#define MID_SQUELCH2    0x21
#define MID_RFP0        0x22
#define MID_RFP6        0x23
#define MID_AGC         0x24
#define MID_TSCN        0x25
#define MID_TANY        0x26
#define MID_TIME        0x27
#define MID_DEEM0       0x28
#define MID_DEEM50      0x29
#define MID_DEEM75      0x2a
#define MID_BKKEEP      0x2b
#define MID_BKADJ       0x2c

#define MID_MODEAM      0x2d
#define MID_MODEFM      0x2e
#define MID_MODEAUX     0x2f

#define MID_FREQ        0x30
#define MID_SEEK        0x31
#define MID_CH          0x32
#define MID_SCAN        0x33
#define MID_ANY         0x34
#define MID_SCSV        0x35

#define MID_LW          0x36
#define MID_MW          0x37
#define MID_SW          0x38
#define MID_SH          0x39
#define MID_FL          0x3a
#define MID_FM          0x3b
#define MID_AL          0x3c
#define MID_AB          0x3d
#define MID_VH          0x3e

#define MID_ALIGNMENT   0x3f

#define MID_FTAM00      100
#define MID_FTAM01      101
#define MID_FTAM02      102
#define MID_FTAM03      103
#define MID_FTAM04      104
#define MID_FTAM05      105
#define MID_FTAM06      106
#define MID_FTAM07      107
#define MID_FTAM08      108
#define MID_FTAM09      109
#define MID_FTAM10      110
#define MID_FTAM11      111
#define MID_FTAM12      112
#define MID_FTAM13      113
#define MID_FTAM14      114
#define MID_FTAM15      115
#define MID_FTAM16      116
#define MID_FTAM17      117
#define MID_FTAM18      118
#define MID_FTAM19      119
#define MID_FTAM20      120
#define MID_FTAM21      121
#define MID_FTAM22      122
#define MID_FTAM23      123
#define MID_FTAM24      124
#define MID_FTAM25      125
#define MID_FTAM26      126
#define MID_FTAM27      127
#define MID_FTAM28      128

#define MID_FTFM00      200
#define MID_FTFM01      201
#define MID_FTFM02      202
#define MID_FTFM03      203
#define MID_FTFM04      204
#define MID_FTFM05      205
#define MID_FTFM06      206
#define MID_FTFM07      207
#define MID_FTFM08      208
#define MID_FTFM09      209
#define MID_FTFM10      210
#define MID_FTFM11      211
#define MID_FTFM12      212
#define MID_FTFM13      213
#define MID_FTFM14      214
#define MID_FTFM15      215
#define MID_FTFM16      216
#define MID_FTFM17      217
#define MID_FTFM18      218
#define MID_FTFM19      219
#define MID_FTFM20      220
#define MID_FTFM21      221
#define MID_FTFM22      222
#define MID_FTFM23      223
#define MID_FTFM24      224
#define MID_FTFM25      225
#define MID_FTFM26      226
#define MID_FTFM27      227
#define MID_FTFM28      228
#define MID_FTFM29      229

// Special menu IDs
#define MID_RET         0xfe
#define MID_EXIT        0xff


// Menu item structure
struct M_ITEM {
	uint8_t nMID;
	const char *pszMTxt;
};

// Sub menu structure
struct M_SUBMENU {
	uint8_t nMID;
	struct M_ITEM *pMItem;
	uint8_t nItemCount;
};


void LCDEN(void);
void LCDCmd(uint8_t Cmd);
void LCDData(uint8_t Data);
void LCDClr(void);
void LCDClr1(void);
void LCDOn(void);
void LCDInit(void);
void LCDXY(uint8_t x, uint8_t y);
void LCDXYChar(uint8_t x, uint8_t y, char c);
void LCDXYStr(uint8_t x, uint8_t y, const char *str);
void LCDXYStr_P(uint8_t x, uint8_t y, PGM_P str);
void LCDXYStrLen(uint8_t x, uint8_t y, char *str, uint8_t nLen, bool bLeftAlign);
void LCDXYStrLen_P(uint8_t x, uint8_t y, PGM_P str, uint8_t nLen, bool bLeftAlign);
void LCDXYIntLen(uint8_t x, uint8_t y, int32_t n, uint8_t nLen);
void LCDXYUIntLenZP(uint8_t x, uint8_t y, uint32_t n, uint8_t nLen);

void intLRotEnc(void);
int8_t GetLRot(void);
void intRRotEnc(void);
int8_t GetRRot(void);
uint8_t PeekKey(void);
uint8_t GetKey(void);
void TestRotKey(void);

void LCDUpdate(void);
int8_t CheckUpdateSig(void);
void ShowMisc(void);
void ShowTime(void);
void ShowVol(void);
void CheckUpdateAlt(int8_t nShow);

void Menu_Squelch(uint8_t nIdx);
void Menu_AGC(void);
void Menu_BacklightAdj(void);
void Menu_BacklightKeep(void);
void Menu_ScanStayTime(void);
void Menu_AnyHoldTime(void);
void Menu_Time(void);
void Menu_Alignment(void);
void Menu_SCSV(void);

bool YesNo(bool bChkSig);
void AddDelCh(void);

void ProcSubMenu(struct M_SUBMENU *pSubMenu);
void ProcMenuItem(uint8_t nMenuID);
void Menu(uint8_t nMenuID);



