/*
 *  Copyright (C) 2012-2014  Konrad Kosmatka, XDR-I2C 2014-02-20
 *
 *  Adapted for YACRC3167 by ace919
 *      V0.0.1  build 7  2014-10-9
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 3
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */


#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include "I2cMaster.h"
#include "global.h"
#include "yacrc3167_ar.h"
#include "init3167.h"
#include "filters.h"
#include "align.h"
#include "ui.h"
#include "nv_memory.h"


extern volatile int8_t nLRot;
extern volatile int8_t nRRot;

extern M_ITEM M_AMFilter[29];
extern M_ITEM M_FMFilter[30];
extern M_ITEM M_Frequency[];
extern M_SUBMENU SM_List[];

extern const uint8_t* const Filters[29];
extern const int8_t AdaptiveFilters[16];
extern const int8_t PrimaryFilters[7];


// TEF6721HL IF
uint16_t PLL;
uint8_t DAA;
uint8_t BAND;
uint8_t AGC = 0x8A;
uint8_t XTAL_DAC = 0xC0;

uint8_t nAGC;

// Modes
uint8_t nMode;    // FM/AM/AUX
uint8_t nRFMode;  // FM/AM


// Volume
uint8_t nVolume;        // Audio volume control, 0-63
bool bMuted = false;    // For mute/unmute


// Tune types
uint8_t nTuneType;  // Freq_Step/Seek/Ch/Scan/Any/Scan_Save

// Bands
uint8_t nBand;  // Band: LW/MW/SW/SH/FL
                //       FM/AL/AB/VH
const uint8_t nBandMode[9] = {MODE_AM, MODE_AM, MODE_AM, MODE_FM, MODE_FM,
                             MODE_FM, MODE_FM, MODE_AM, MODE_FM};  // Band default RF mode
extern const int32_t nBandFMin[9] = {1, 531, 1711, 19305, 64000,
                                     87500, 108005, 117975, 137010};
extern const int32_t nBandFMax[9] = {530, 1710, 19300, 63995, 87490,
                                     108000, 117970, 137000, 289300};
extern const int16_t nBandStep[9][3] = { {9, 1, 45}, {9, 1, 90}, {5, 1, 500}, {5, 50, 500}, {10, 50, 500},
                                         {100, 5, 500}, {100, 5, 500}, {25, 10, 500}, {25, 10, 500} };
extern const uint8_t nBandChs[9] = {CHS_LW, CHS_MW, CHS_SW, CHS_SH, CHS_FL,
                                    CHS_FM, CHS_AL, CHS_AB, CHS_VH};  // Band total channels
extern const uint16_t nBandChNVMA[9] = {NVMADDR_LW, NVMADDR_MW, NVMADDR_SW, NVMADDR_SH, NVMADDR_FL,
                                        NVMADDR_FM, NVMADDR_AL, NVMADDR_AB, NVMADDR_VH};
int32_t nBandFreq[9];  // Band default frequency
int8_t nBandCh[9];     // Band default channel No.


// Steps
uint8_t nStepIdx;   // Step index for current band


// Filters
int8_t nAMFilter;   // Current FIR filter index for AM
int8_t nFMFilter;   // Current FIR filter index for FM(ADP_FM_FILTER is adaptive)

uint8_t nDeemphasis;

int8_t nSquelch[2];  // Signal squelch value in dBuv, -99~99
                     // 1st element for auto mute, 2nd element for seek/scan/any
//bool bSquelched = false;

uint8_t nBacklightAdj;         // LCD backlight value, 0-255
uint8_t nBacklightKeep;        // LCD backlight auto keep seconds, 0-255, 0 for always on
bool bLCDOff = false;          // true for LCD is off
uint32_t nBacklightTimer;      // LCD backlight auto keep timer in ms

uint32_t timer = 0;   // Signal level & FM stereo indicator reporting timer
int32_t nSecondsOffset = 3600L * 24 * 15;  // Seconds of real time offset, preset to 15 days

uint8_t nScanStayTime;  // Seconds to stay at current frequency
uint8_t nAnyHoldTime;   // Seconds to hold current frequency when lost signal

uint16_t nAutoSyncBits;
uint16_t nAutoSyncChs;
uint16_t nAutoSyncFreqs;
uint32_t nAutoSyncTimer;  // Auto sync to NV memory timer
bool bBandChange = false;

TwiMaster i2c(false);



////////////////////////////////////////////////////////////
void dsp_start_addr(uint32_t addr)
{
    i2c.start(DSP_I2C | I2C_WRITE);
    i2c.write((uint8_t)(addr >> 16));
    i2c.write((uint8_t)(addr >> 8));
    i2c.write((uint8_t)addr);
}

uint8_t dsp_query1(uint32_t addr)
{
    dsp_start_addr(addr);
    i2c.restart(DSP_I2C | I2C_READ);
    uint8_t buffer = i2c.read(true);
    i2c.stop();
    return buffer;
}

uint16_t dsp_query2(uint32_t addr)
{
    dsp_start_addr(addr);
    i2c.restart(DSP_I2C | I2C_READ);
    uint16_t buffer = ((uint16_t)i2c.read(false) << 8);
    buffer |= i2c.read(true);
    i2c.stop();
    return buffer;
}

void dsp_write2(uint32_t addr, uint16_t data)
{
    dsp_start_addr(addr);
    i2c.write((uint8_t)(data >> 8));
    i2c.write((uint8_t)data);
    i2c.stop();
}

void dsp_write3(uint32_t addr, uint32_t data)
{
    dsp_start_addr(addr);
    i2c.write((uint8_t)(data >> 16));
    i2c.write((uint8_t)(data >> 8));
    i2c.write((uint8_t)data);
    i2c.stop();
}

void dsp_write_data(const uint8_t* data)
{
    uint8_t *pa = (uint8_t *)data;
    uint8_t len, i;

    for(;;) {
        len = pgm_read_byte_near(pa++);
        if(!len)
            break;
        if(len & 0x80)  // Delay
            delay( ((uint16_t)(len & 0x7f) << 8 ) | \
                pgm_read_byte_near(pa++) );
        else {
            i2c.start(DSP_I2C | I2C_WRITE);
            for(i = 0; i < len; i++)
                i2c.write(pgm_read_byte_near(pa++));
            i2c.stop();
        }
    }
}

void dsp_write_data_len(const uint8_t* data, uint32_t addr, uint8_t len)
{
    uint8_t *pa = (uint8_t *)data;
    uint8_t i;

    dsp_start_addr(addr);
    for(i = 0; i < len; i++)
        i2c.write(pgm_read_byte_near(pa++));
    i2c.stop();
}

void dsp_write_coeff(uint8_t bank, uint8_t filter)
{
    uint8_t i = 0;
    uint32_t address = 0x010C00 + 32 * bank;

    while(i < 64) {
        dsp_start_addr(address);
        i2c.write(pgm_read_byte_near(Filters[filter] + (i++)));
        i2c.write(pgm_read_byte_near(Filters[filter] + (i++)));
        i2c.write(0x00);
        i2c.stop();
        address++;
    }
}


////////////////////////////////////////////////////////////
void SetVolume(uint8_t v64)
{
    uint16_t v, v0, LdComp;
    static uint8_t vLast = 0;

    if(v64 == vLast)
        return;

    v = v64 * 32;
    if(nMode == MODE_AM) {
        v0 = v;
        if(v >= 0x047F)
            LdComp = 0x0800;
        else if(v >= 0x04B8)
            LdComp = 0x0935;
        else if(v >= 0x04AF)
            LdComp = 0x09E6;
        else if(v >= 0x04A4)
            LdComp = 0x0A83;
        else if(v >= 0x0499)
            LdComp = 0x0B10;
        else if(v >= 0x0410)
            LdComp = 0x0B91;
        else if(v >= 0x03EE)
            LdComp = 0x0BEA;
        else if(v >= 0x03E1)
            LdComp = 0x0C50;
        else if(v >= 0x03D5)
            LdComp = 0x0CAC;
        else if(v >= 0x0301)
            LdComp = 0x0CFF;
        else if(v >= 0x02F8)
            LdComp = 0x0D4A;
        else if(v >= 0x02EE)
            LdComp = 0x0D8D;
        else if(v >= 0x024C)
            LdComp = 0x0DCA;
        else if(v >= 0x0206)
            LdComp = 0x0E02;
        else if(v >= 0x0169)
            LdComp = 0x0E33;
        else if(v >= 0x0165)
            LdComp = 0x0E60;
        else if(v >= 0x00DE)
            LdComp = 0x0E88;
        else if(v >= 0x008A)
            LdComp = 0x0EAD;
        else if(v >= 0x0008)
            LdComp = 0x0ECE;
        else
            LdComp = 0x0FFB;
    }

    else {  // FM/AUX mode
        if(v64 >= 22)
            LdComp = 0x0800;
        else if(v64 >= 20)
            LdComp = 0x0899;
        else if(v64 >= 2)
            LdComp = 0x095D;
        else if(v64 >= 1)
            LdComp = 0x0C51;
        else
            LdComp = 0x0FAC;

        if(v64 >= 63)
            v0 = 0x07FF;
        else if(v64 >= 61)
            v0 = 0x0721;
        else if(v64 >= 59)
            v0 = 0x065A;
        else if(v64 >= 57)
            v0 = 0x05A9;
        else if(v64 >= 55)
            v0 = 0x050C;
        else if(v64 >= 39)
            v0 = 0x07FF;
        else  // v64=0-38
            v0 =  ( (uint16_t)pgm_read_byte_near(VOL_V0_LIST + (v64 << 1)) << 8 ) |
                     pgm_read_byte_near(VOL_V0_LIST + (v64 << 1) + 1);

        uint16_t v2;
        if(v64 >= 55)
            v2 = 0x0800;
        else if(v64 <= 38)
            v2 = 0x0EFF;
        else  // v64=39-54
            v2 =  ( (uint16_t)pgm_read_byte_near(VOL_V2_LIST + ((v64 - 39) << 1)) << 8 ) |
                     pgm_read_byte_near(VOL_V2_LIST + ((v64 - 39) << 1) + 1);

        dsp_write2(DSP_VOLUME_SCALER2, v2);
    }

    dsp_write2(DSP_LOUDNESS_COMP, LdComp);
    dsp_write2(DSP_VOLUME_SCALER, v0);
    vLast = v64;
}


void CheckVolume(void)
{
    int8_t i8;

    if( (i8 = GetLRot()) ) {
        if( (i8 < 0) && (-i8 > nVolume) )
            nVolume = 0;
        else
            nVolume = constrain(nVolume + i8, MIN_VOL, MAX_VOL);
        SetVolume(nVolume);
        if(nVolume)
            bMuted = false;
        AddSyncBits(NEEDSYNC_VOL);
        CheckUpdateAlt(ALT_VOL);  // Show volume for a period
    }
}


void dsp_set_filter(int8_t f)
{
    if(f < ADP_FM_FILTER) {  // fixed filter bandwidth
        if(nRFMode == MODE_AM) {
            // workaround for AM (?)
            for(uint8_t i = 0; i < 16; i++)
                dsp_write_coeff(i, f);
//            AddSyncBits(NEEDSYNC_AMFILTER);
            return;
        }

        // write the FIR filter coefficients into #15 filter bank
        dsp_write_coeff(15, f);

        dsp_write3(TDSP1_X_CIBW_1_FirCtlFix, 0x00000F);  // #15 filter
        dsp_write3(TDSP1_X_CIBW_4_FirCtlFix, 0x00000F);  // #15 filter

        dsp_write3(TDSP1_X_CIBW_1_pFirCtl, (uint16_t)TDSP1_X_CIBW_1_FirCtlFix);  // relative address!
        dsp_write3(TDSP1_X_CIBW_4_pFirCtl, (uint16_t)TDSP1_X_CIBW_4_FirCtlFix);  // relative address!
    }
    else if(nRFMode == MODE_FM) {  // Adaptive filter bandwidth for FM
        for(uint8_t i = 0; i < 16; i++)
            dsp_write_coeff(i, AdaptiveFilters[i]);

        dsp_write3(TDSP1_X_CIBW_1_pFirCtl, (uint16_t)TDSP1_X_CIBW_1_FirCtl);  // relative address!
        dsp_write3(TDSP1_X_CIBW_4_pFirCtl, (uint16_t)TDSP1_X_CIBW_4_FirCtl);  // relative address!
    }

//    AddSyncBits(NEEDSYNC_FMFILTER);
}


void NextPriFilter(void)
{
    uint8_t i, CurrIdx;

    if(nRFMode == MODE_FM) {
        if(nFMFilter == ADP_FM_FILTER) {
             nFMFilter = PrimaryFilters[0];
             return;
        }
        CurrIdx = nFMFilter;
    }
    else
        CurrIdx = nAMFilter;

    for(i = 0; i < sizeof(PrimaryFilters); i++) {
        if(CurrIdx <= PrimaryFilters[i]) {
            CurrIdx = PrimaryFilters[i];
            break;
        }
    }

    if(nRFMode == MODE_FM) {
        if(CurrIdx >= PrimaryFilters[sizeof(PrimaryFilters) - 1])
            nFMFilter = ADP_FM_FILTER;
        else
            nFMFilter = PrimaryFilters[i + 1];
    }
    else {  // MODE_AM
        if(CurrIdx >= PrimaryFilters[sizeof(PrimaryFilters) - 1])
            nAMFilter = PrimaryFilters[0];
        else
            nAMFilter = PrimaryFilters[i + 1];
    }
}


void SetFilter(bool bNeedSync)
{
    if(nRFMode == MODE_FM) {
        dsp_set_filter(nFMFilter);
        if(bNeedSync)
            AddSyncBits(NEEDSYNC_FMFILTER);
    }
    else {
        dsp_set_filter(nAMFilter);
        if(bNeedSync)
            AddSyncBits(NEEDSYNC_AMFILTER);
    }
}


void dsp_set_deemphasis(uint8_t d)
{
    if(d <= 2) {
        nDeemphasis = d;
        dsp_write_data(deemph[d]);
    }
}


void TuneReg(void)
{
    i2c.SetBitRate(83333L);  // Set I2C bit rate to 83.333KHz for TEF6721
    for(int8_t i = 1; i >= 0; i--) {
        dsp_start_addr(0x00FFFF);
        i2c.restart(IF_I2C | I2C_WRITE);
        i2c.write(PLL >> 8);
        i2c.write(PLL & 0xFF);
        i2c.write(DAA | (i << 7));
        i2c.write(BAND);
        if(nRFMode == MODE_AM)
            i2c.write(AGC | (nAGC << 5));
        else
            i2c.write(AGC | ((3 - nAGC) << 5));
        i2c.write(XTAL_DAC);
        i2c.stop();
    }
    i2c.SetBitRate(F_TWI);  // Resume default I2C bit rate for SAF7730
}

bool TuneFreq(int32_t freq)
{
    bool bFMLowInj = false;

    if(freq >= 1 && freq <= 2100) {  // 0.001-2.1MHz. LW/MW(extended)
        BAND = B01100101;  // div=20, fref=20KHz, 1KHz step
        PLL = freq + 10700;
    }
    else if(freq > 2100 && freq <= 19300) {  // 2.1-19.3MHz. SW
        BAND = B00010001;  // div=10, fref=10KHz, 1KHz step
        PLL = freq + 10700;
    }
    else if( (freq > 19300 && freq < 64000) || (freq >= 87500 && freq < 117975) ) {
        // 19.3-64MHz, 87.5-117.975MHz. SW/FM(extended)
        if(freq % 50) {
            BAND = B00010000;  // div=2, fref=10KHz, 5KHz step
            PLL = (freq + 10700 + 2) / 5;
        }
        else {
            // We use 50KHz step if possible
            BAND = B00000000;  // div=2, fref=100KHz, 50KHz step
            PLL = (freq + 10700) / 50;
        }
    }
    else if(freq >= 64000 && freq < 87500) {  // 64-87.5MHz. FM east Europe, FM Japan
        if(freq % 100) {
            BAND = B00010010;  // div=3, fref=10KHz, 3.333KHz(x3=10k) step
            PLL = ((freq - 10700) * 3 + 5) / 10;
        }
        else {
            // We use 33.333KHz step if possible
            BAND = B00000010;  // div=3, fref=100KHz, 33.333KHz(x3=100k) step
            PLL = (freq - 10700) * 3 / 100;
        }
        bFMLowInj = true;  // FM low injection
    }
    else if(freq >= 117975 && freq <= 289300) {  // 117.975-289.3MHz. Air band, VHF
        if(freq % 25) {
            BAND = B00010110;  // div=1, fref=10KHz, 10KHz step
            PLL = (freq + 10700 + 5) / 10;
        }
        else {
            // We use 25KHz step if possible
            BAND = B00100110;  // div=1, fref=25KHz, 25KHz step
            PLL = (freq + 10700) / 25;
        }
    }
    else
        return false;

    if(bFMLowInj)
        AGC &= B11110111;  // FM low injection
    else
        AGC |= B00001000;  // FM high injection
    SetDAA(freq);
    TuneReg();
    if(!bBandChange) {
        nAutoSyncFreqs |= (1 << nBand);
        nAutoSyncTimer = millis();
    }
    return true;
}


void AdjFreq(void)
{
    if(nBandFreq[nBand] < nBandFMin[nBand])
        nBandFreq[nBand] = nBandFMax[nBand];
    else if(nBandFreq[nBand] > nBandFMax[nBand])
        nBandFreq[nBand] = nBandFMin[nBand];

    // Adjust frequency to integer multiples of step value
    nBandFreq[nBand] = (nBandFreq[nBand] / nBandStep[nBand][nStepIdx]) * nBandStep[nBand][nStepIdx];

    if(nBandFreq[nBand] <= 0)  // For LW only
        nBandFreq[nBand] = nBandFMin[nBand];
}


void TuneFreqDisp(void)
{
    TuneFreq(nBandFreq[nBand]);
    LCDXYIntLen(FREQ_X, FREQ_Y, nBandFreq[nBand], 6);
}


uint32_t ReadChFreq(void)
{
    uint16_t nNVMData;
    int32_t nFreq;

    nNVMData = NV_read_word((uint16_t *)(nBandChNVMA[nBand] + nBandCh[nBand] * 2));
    if(nNVMData == 0xffff)  // Blank ch
        return 0;
    if(nBand <= BAND_SW)
        nFreq = nBandFMin[nBand] + nNVMData;
    else
        nFreq = nBandFMin[nBand] + (int32_t)nNVMData * 5;
    nFreq = constrain(nFreq, nBandFMin[nBand], nBandFMax[nBand]);

    return nFreq;
}


void WriteChFreq(bool bAdd)
{
    uint16_t u16;

    if(bAdd) {  // Add ch
        if(nBand <= BAND_SW)
            u16 = (uint16_t)(nBandFreq[nBand] - nBandFMin[nBand]);
        else
            u16 = (uint16_t)((nBandFreq[nBand] - nBandFMin[nBand]) / 5);
    }
    else  // Delete ch
        u16 = 0xffff;

    NV_write_word( (uint16_t *)(nBandChNVMA[nBand] + nBandCh[nBand] * 2), u16);
}


void SeekCh(int8_t nDir)
{
    uint8_t u8;
    uint32_t u32;
    int8_t nDir1;

    nDir1 = nDir;
    if(!nDir1) {  // Try current ch 1st
        if( (u32 = ReadChFreq()) ) {
            nBandFreq[nBand] = u32;
            TuneFreqDisp();
            LCDUpdate();
            return;
        }
        nDir1 = 1;
    }

    for(u8 = 0; u8 < nBandChs[nBand]; u8++) {
        nBandCh[nBand] = (nBandCh[nBand] + nDir1 + nBandChs[nBand]) % nBandChs[nBand];
        if( (u32 = ReadChFreq()) ) {
            nBandFreq[nBand] = u32;
            TuneFreqDisp();
            LCDUpdate();
            nAutoSyncChs |= (1 << nBand);
            nAutoSyncTimer = millis();
            return;
        }
    }

    // No ch find, set ch=0, f=fmin
    nBandCh[nBand] = 0;
    nBandFreq[nBand] = nBandFMin[nBand];
    TuneFreqDisp();
    LCDUpdate();
}


int8_t GetdBuv(void)
{
    if(nRFMode == MODE_FM)
        return dsp_query1(DSP_FM_LEVEL) - 11;
    else
        return dsp_query1(DSP_AM_LEVEL) - 11;
}


void ProcBand(uint8_t nBd)
{
    if(nStepIdx) {
        nStepIdx = 0;  // Default band step
        AddSyncBits(NEEDSYNC_STEPIDX);
    }

    if( (nBandMode[nBand] != nBandMode[nBd]) || (nMode == MODE_AUX) ) {
        nBand = nBd;
        SetRFMode();
    }
    nBand = nBd;

    if(nTuneType == TYPE_CH)
        SeekCh(0);
    else {
        bBandChange = true;  // Change band only, don't sync band frequency to NV memory
        TuneFreqDisp();
        bBandChange = false;
    }

    AddSyncBits(NEEDSYNC_BAND);
}


void ProcStepFilter(uint8_t nKey)
{
    switch(nKey) {
    case KEY_STEP:
        nStepIdx = (nStepIdx + 1) % NUM_STEPS;
        AddSyncBits(NEEDSYNC_STEPIDX);
        break;

    case KEY_STEP | KEY_LONGPRESS:   // Default step
        if(nStepIdx != 0)
            nStepIdx = 0;
        else
            nStepIdx = NUM_STEPS - 1;
        AddSyncBits(NEEDSYNC_STEPIDX);
        break;

    case KEY_FILTER:
        NextPriFilter();
        SetFilter(true);
        break;

    case KEY_FILTER | KEY_LONGPRESS:  // Default filter
        if(nRFMode == MODE_FM) {
            if(nFMFilter != DEF_FM_FILTER)
                nFMFilter = DEF_FM_FILTER;
            else
                nFMFilter = DEF_FM_FILTER2;
        }
        else {
            if(nAMFilter != DEF_AM_FILTER)
                nAMFilter = DEF_AM_FILTER;
            else
                nAMFilter = DEF_AM_FILTER2;
        }
        SetFilter(true);
        break;
    }
}


int8_t Seek(int8_t nDir)
{
    int8_t i;
    uint8_t nKey;

    SetVolume(0);  // Mute
    if(nRFMode == MODE_AM) {
        if(nAMFilter > SEEK_AM_FILTER)
            dsp_set_filter(SEEK_AM_FILTER);
    }
    else {
        if(nFMFilter > SEEK_FM_FILTER)
            dsp_set_filter(SEEK_FM_FILTER);
    }
    for(;;) {
        nBandFreq[nBand] += nDir * nBandStep[nBand][nStepIdx];
        AdjFreq();
        TuneFreqDisp();
        if(nRFMode == MODE_AM)
            delay(25);
        else
            delay(5);

        for(i = 0; i < 30; i++) {
            if( (nKey = GetKey()) ) {
                if(nKey & (KEY_STEP | KEY_FILTER)) {
                    ProcStepFilter(nKey);
                    LCDUpdate();
                }
                else {
                    SetFilter(false);
                    SetVolume(nVolume);  // Unmute
                    return nKey;  // Other key pressed, abort seek
                }
            }

            delay(1);
            if(GetdBuv() >= nSquelch[1]) {  // Find one signal
                SetFilter(false);
                SetVolume(nVolume);  // Unmute
                return 0;
            }
        }

        CheckUpdateSig();
        CheckVolume();
        CheckUpdateAlt(ALT_AUTO);
    }
}


uint8_t ScanAny()
{
    uint32_t t1, nTimeout;
    int8_t i8;
    uint8_t nKey;

    if(nTuneType == TYPE_SCAN)
        nTimeout = nScanStayTime * 1000;
    else
        nTimeout = nAnyHoldTime * 1000;
    LCDUpdate();
    NV_write_byte((uint8_t *)NVMADDR_TUNE, nTuneType);

    for(;;) {
        if( (nKey = Seek(+1)) )  // Key(Other than STEP & FILTER) pressed, user abort
            return nKey;

        LCDOn();
        // Else we  find one signal
        t1 = millis();
        for(;;) {
            i8 = CheckUpdateSig();;
            if(nTuneType == TYPE_ANY && i8 >= nSquelch[1])
                t1 = millis();

            if( (millis() - t1) > nTimeout )  // Lost signal for nTimeout ms in ANY mode
                break;                        // or nTimeout ms passed in SCAN mode

            if( (nKey = GetKey()) ) {
                if(nKey & (KEY_STEP | KEY_FILTER)) {
                    ProcStepFilter(nKey);
                    LCDUpdate();
                }
                else
                    return nKey;  // Key(Other than STEP/FILTER) pressed, user abort
            }

            if(GetRRot())  // Right rotary encoder turned, leave this signal
                break;

            if((nTuneType == TYPE_ANY ) && nBacklightKeep && !bLCDOff) {
                if( (millis() - nBacklightTimer ) >= (uint32_t)nBacklightKeep * 1000 ) {
                    analogWrite(LCD_BKLT_PIN, 0);  // Set LCD backlight off
                    bLCDOff = true;
                }
            }

            CheckVolume();  // Check volume adjustment
            CheckUpdateAlt(ALT_AUTO);
            delay(200);
        }
    }
}


void SetModeAM(void)
{
    nRFMode = nMode = MODE_AM;
    SM_List[MID_FILT].pMItem = M_AMFilter;
    SM_List[MID_FILT].nItemCount = sizeof(M_AMFilter) / sizeof(M_ITEM);

    dsp_write_data(DSP_AM_HEAD);
    dsp_write_data_len(DSP_AM_MID, 0x0D1192, sizeof(DSP_AM_MID));
    dsp_write_data_len(DSP_AM_MID, 0x0D11B0, sizeof(DSP_AM_MID));
    dsp_write_data_len(DSP_AM_MID, 0x0D11CE, sizeof(DSP_AM_MID));
    dsp_write_data_len(DSP_AM_MID, 0x0D11EC, sizeof(DSP_AM_MID));
    dsp_set_filter(nAMFilter);
    dsp_write_data(DSP_AM_TAIL);
    AddSyncBits(NEEDSYNC_MODE | NEEDSYNC_RFMODE);
}

void SetModeFM(void)
{
    nRFMode = nMode = MODE_FM;
    SM_List[MID_FILT].pMItem = M_FMFilter;
    SM_List[MID_FILT].nItemCount = sizeof(M_FMFilter) / sizeof(M_ITEM);

    dsp_write_data(DSP_FM_HEAD);
    dsp_write_data_len(DSP_FM_MID, 0x0D1192, sizeof(DSP_FM_MID));
    dsp_write_data_len(DSP_FM_MID, 0x0D11B0, sizeof(DSP_FM_MID));
    dsp_write_data_len(DSP_FM_MID, 0x0D11CE, sizeof(DSP_FM_MID));
    dsp_write_data_len(DSP_FM_MID, 0x0D11EC, sizeof(DSP_FM_MID));
    dsp_set_filter(nFMFilter);
    dsp_write_data(DSP_FM_TAIL);
    AddSyncBits(NEEDSYNC_MODE | NEEDSYNC_RFMODE);
}

void SetModeAUX(void)
{
    nMode = MODE_AUX;
    dsp_write_data(DSP_AUX_HEAD);
    dsp_write_data_len(DSP_AUX_MID1, 0x0D1192, sizeof(DSP_AUX_MID1));
    dsp_write_data_len(DSP_AUX_MID1, 0x0D11B0, sizeof(DSP_AUX_MID1));
    dsp_write_data_len(DSP_AUX_MID2, 0x0D11CE, sizeof(DSP_AUX_MID2));
    dsp_write_data_len(DSP_AUX_MID2, 0x0D11EC, sizeof(DSP_AUX_MID2));
    dsp_write_data(DSP_AUX_TAIL);
    AddSyncBits(NEEDSYNC_MODE);
}

void SetRFMode(void)
{
    if(nBandMode[nBand] == MODE_FM)
        SetModeFM();
    else
        SetModeAM();
}


void AddSyncBits(uint16_t SyncBit)
{
    nAutoSyncBits |= SyncBit;
    nAutoSyncTimer = millis();
}



////////////////////////////////////////////////////////////
void setup(void)
{
    pinMode(SDA_PIN, INPUT);
    pinMode(SCL_PIN, INPUT);
    pinMode(RDS_PIN, INPUT);
    pinMode(RST_PIN, OUTPUT);
    digitalWrite(RST_PIN, LOW);
    pinMode(SH1A_PIN, INPUT_PULLUP);
    pinMode(SH1B_PIN, INPUT_PULLUP);
    pinMode(SH2A_PIN, INPUT_PULLUP);
    pinMode(SH2B_PIN, INPUT_PULLUP);

#if DEBUG
    Serial.begin(115200);
#endif

    LCDInit();
    LCDXYStr_P(4, 0, (PGM_P)F("YACRC3167"));
    LCDXYStr_P(0, 1, (PGM_P)F("V0.0.1 Build 7"));

    delay(1000);
    digitalWrite(RST_PIN, HIGH);
    delay(30);
    digitalWrite(RST_PIN, LOW);
    delay(30);

    attachInterrupt(0, intLRotEnc, CHANGE);
    attachInterrupt(1, intRRotEnc, CHANGE);
    if(analogRead(0) < ADC_NONE / 2)
        TestRotKey();

    digitalWrite(RST_PIN, HIGH);
    delay(2000);
    dsp_write_data(DSP_INIT);
    NVMGetArgs();
    LCDClr();
    nLRot = 0;
    nRRot = 0;

    SetRFMode();            // Set RF mode
    if(nTuneType == TYPE_CH)
        SeekCh(0);          // Tune to current ch
    else
        TuneFreqDisp();     // Tune to default frequency and display
    SetVolume(nVolume);     // Set audio volume
    dsp_set_deemphasis(nDeemphasis);
    analogWrite(LCD_BKLT_PIN, nBacklightAdj);  // Set LCD backlight PWM
    nBacklightTimer = millis();
    LCDUpdate();

    nAutoSyncBits = 0;
    nAutoSyncChs = 0;
    nAutoSyncFreqs = 0;
    nAutoSyncTimer = 1000L * 3600 * 24;  // Auto sync to NV memory timer

#if DEBUG
    Serial.println(F("End setup"));
#endif
}


////////////////////////////////////////////////////////////
void loop(void)
{
    int8_t i8;
    uint8_t nKey;
    uint16_t u16;


    // Check signal level and 19kHz subcarrier every TIMER_INTERVAL
    if( (millis() - timer ) >= TIMER_INTERVAL) {
        i8 = CheckUpdateSig();  // Update RF signal level in dBuv & FM stereo indicator
        timer = millis();

        if(!bMuted && (nMode != MODE_AUX)) {
            if( (i8 < nSquelch[0]) )
                SetVolume(0);
            else
                SetVolume(nVolume);
        }

        CheckUpdateAlt(ALT_AUTO);

        if(nBacklightKeep && !bLCDOff) {
            if( (timer - nBacklightTimer ) >= (uint32_t)nBacklightKeep * 1000 ) {
                analogWrite(LCD_BKLT_PIN, 0);  // Set LCD backlight off
                bLCDOff = true;
            }
        }

        if( (timer - nAutoSyncTimer) >= TIMER_AUTOSYNC ) {
            if(nAutoSyncBits) {
                if(nAutoSyncBits & NEEDSYNC_BAND)
                    NV_write_byte((uint8_t *)NVMADDR_BAND, nBand);

                if(nAutoSyncBits & NEEDSYNC_VOL)
                    NV_write_byte((uint8_t *)NVMADDR_VOL, nVolume);

                if(nAutoSyncBits & NEEDSYNC_MODE)
                    NV_write_byte((uint8_t *)NVMADDR_MODE, nMode);

                if(nAutoSyncBits & NEEDSYNC_RFMODE)
                    NV_write_byte((uint8_t *)NVMADDR_RFMODE, nRFMode);

                if(nAutoSyncBits & NEEDSYNC_TUNE)
                    NV_write_byte((uint8_t *)NVMADDR_TUNE, nTuneType);

                if(nAutoSyncBits & NEEDSYNC_STEPIDX)
                    NV_write_byte((uint8_t *)NVMADDR_STEPIDX, nStepIdx);

                if(nAutoSyncBits & NEEDSYNC_AMFILTER)
                    NV_write_byte((uint8_t *)NVMADDR_AMFILTER, nAMFilter);

                if(nAutoSyncBits & NEEDSYNC_FMFILTER)
                    NV_write_byte((uint8_t *)NVMADDR_FMFILTER, nFMFilter);

                if(nAutoSyncBits & NEEDSYNC_MISC1)
                    NV_write_byte((uint8_t *)NVMADDR_MISC1, (nDeemphasis << 6) | ((AGC & 1) << 5) | nAGC);

                nAutoSyncBits = 0;
            }

            if(nAutoSyncChs) {
	            for(i8 = 0; i8 < NUM_BANDS; i8++) {
	                if( nAutoSyncChs & (1 << i8) )
	                    NV_write_byte( (uint8_t *)NVMADDR_BANDCH + i8, nBandCh[i8]);
                }
                nAutoSyncChs = 0;
            }

            if(nAutoSyncFreqs) {
	            for(i8 = 0; i8 < NUM_BANDS; i8++) {
	                if( nAutoSyncFreqs & (1 << i8) ) {
	                    if(i8 <= BAND_SW)
                            u16 = (uint16_t)(nBandFreq[i8] - nBandFMin[i8]);
                        else
                            u16 = (uint16_t)((nBandFreq[i8] - nBandFMin[i8]) / 5);
	                    NV_write_word( (uint16_t *)NVMADDR_BANDFREQ + i8, u16);
	                }
                }
                nAutoSyncFreqs = 0;
            }

            nAutoSyncTimer = timer + 1000L * 3600 * 24;
        }
    }


    CheckVolume();


    if( (i8 = GetRRot()) ) {
        switch(nTuneType) {
        case TYPE_FREQ:
            nBandFreq[nBand] += i8 * nBandStep[nBand][nStepIdx];
            AdjFreq();
            TuneFreqDisp();
            break;

        case TYPE_SEEK:
            Seek(constrain(i8, -1, 1));
            break;

        case TYPE_CH:
            if(i8 < 0)
                SeekCh(-1);
            else
                SeekCh(1);
            break;
        }
    }


    if( (nKey = GetKey()) ) {
        switch(nKey) {
        case KEY_LROT:
            Menu(MID_OPTION);
            break;

        case KEY_LROT | KEY_LONGPRESS:  // Toggle mute/unmute
            if(bMuted)
                SetVolume(nVolume);  // Unmute
            else
                SetVolume(0);  // Mute
            bMuted = !bMuted;
            break;

        case KEY_RROT:
            Menu(MID_FREQUENCY);
            break;

        case KEY_RROT | KEY_LONGPRESS:
            AddDelCh();
            LCDUpdate();
            break;

        case KEY_TUNE:
            nTuneType = (nTuneType + 1) % NUM_TYPES;
            if(nTuneType == TYPE_CH)
                SeekCh(0);
            AddSyncBits(NEEDSYNC_TUNE);
            break;

        case KEY_TUNE | KEY_LONGPRESS:
            if(nTuneType != TYPE_FREQ)
                nTuneType = TYPE_FREQ;
            else {
                nTuneType = TYPE_CH;
                SeekCh(0);
            }
            AddSyncBits(NEEDSYNC_TUNE);
            break;

        case KEY_BAND:
            ProcBand((nBand + 1) % NUM_BANDS);
            break;

        case KEY_BAND | KEY_LONGPRESS:
            if(nBand != BAND_FM)
                ProcBand(BAND_FM);
            else
                ProcBand(BAND_MW);
            break;

        default:  // Include KEY_STEP, KEY_STEP | KEY_LONGPRESS, KEY_FILTER, KEY_FILTER | KEY_LONGPRESS
            ProcStepFilter(nKey);
            break;
        }
        LCDUpdate();
   }

    if(nTuneType == TYPE_SCAN) {
        if( ScanAny() == KEY_TUNE )
            nTuneType = TYPE_ANY;
        else
           nTuneType = TYPE_FREQ;
        LCDUpdate();
        AddSyncBits(NEEDSYNC_TUNE);
    }

    if(nTuneType == TYPE_ANY) {
        ScanAny();
        nTuneType = TYPE_FREQ;
        LCDUpdate();
        AddSyncBits(NEEDSYNC_TUNE);
    }
}
